// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#include <dolfin/fem/assemble.h>
#include <dolfin/la/solve.h>
#include <dolfin/parameter/parameters.h>
#include "dolfin/PlasticityProblem.h"
#include "../../forms/3Dlinear_QE_forms/Plas3D.h"
#include "../../forms/3Dlinear_QE_forms/Strain3D.h"
#include "../../forms/3Dlinear_QE_forms/p_strain3D.h"
#include "../../forms/3Dlinear_QE_forms/ep_strain3D.h"

using namespace dolfin;

//-----------------------------------------------------------------------------
// Constructor
PlasticityProblem::PlasticityProblem(Function& u, Function& b, Mesh& mesh, 
          Array<DirichletBC*>& bcs, bool& elastic_tangent, PlasticityModel& plastic_model):
          NonlinearProblem(), assembler(mesh), _mesh(&mesh), _bcs(bcs),
          _elastic_tangent(&elastic_tangent), _plastic_model(&plastic_model), reset_tensor(true)
{

  // Init forms
  a = new Plas3DBilinearForm(consistent_tangent_function);
  a_strain = new Strain3DBilinearForm;
  a_plastic_strain = new p_strain3DBilinearForm;
  a_equivalent_plastic_strain = new ep_strain3DBilinearForm;
  L = new Plas3DLinearForm(b, stress_function);
  L_strain = new Strain3DLinearForm(u);

  // Assemble LHS to compute strains, only needs to be computed once
  assemble(A_strain, *a_strain, mesh);

  // Initialise solution function
  u.init(mesh, *a, 0);

  // Initialise other functions and reference their vectors
  plastic_strain_old_function.init(mesh, *a_plastic_strain, 0);
  plastic_strain_old = &plastic_strain_old_function.vector();

  plastic_strain_new_function.init(mesh, *a_plastic_strain, 0);
  plastic_strain_new = &plastic_strain_new_function.vector();

  consistent_tangent_function.init(mesh, *a, 2);
  consistent_tangent_vec = &consistent_tangent_function.vector();

  equivalent_plastic_strain_old_function.init(mesh, *a_equivalent_plastic_strain);
  equivalent_plastic_strain_old = &equivalent_plastic_strain_old_function.vector();

  equivalent_plastic_strain_new_function.init(mesh, *a_equivalent_plastic_strain);
  equivalent_plastic_strain_new = &equivalent_plastic_strain_new_function.vector();

  strain_function.init(mesh, *L_strain, 0);
  strain = &strain_function.vector();

  stress_function.init(mesh, *L_strain, 0);
  stress = &stress_function.vector();
}
//-----------------------------------------------------------------------------
// Destructor
PlasticityProblem::~PlasticityProblem()
{
   delete a; delete a_strain; delete a_plastic_strain; delete a_equivalent_plastic_strain;
   delete L; delete L_strain;
}
//-----------------------------------------------------------------------------
// User defined residual vector 
void PlasticityProblem::F(GenericVector& b, const GenericVector& x)
{
  // Only call form once to loop elements and determine plastic set
  form();

  dolfin_set("output destination", "silent");

  // Assemble RHS
  assembler.assemble(b, *L);

  // Apply boundary conditions
  for (dolfin::uint i(0); i < _bcs.size(); ++i)
    _bcs[i]->apply(b, x, *a);

  dolfin_set("output destination", "terminal");
}
//-----------------------------------------------------------------------------
// User defined assemble of Jacobian 
void PlasticityProblem::J(GenericMatrix& A, const GenericVector& x)
{
  dolfin_set("output destination", "silent");

  // Assemble LHS
  assembler.assemble(A, *a, reset_tensor);
  reset_tensor = false;

  // Apply boundary conditions
  for (dolfin::uint i(0); i < _bcs.size(); ++i)
    _bcs[i]->apply(A, *a);
  
  dolfin_set("output destination", "terminal");
}
//-----------------------------------------------------------------------------
// Loop integration points and compute stresses and consistent tangent
void PlasticityProblem::form()
{
  // Compute strains
  dolfin_set("output destination", "silent");
  assembler.assemble(b_strain, *L_strain);
  solve(A_strain, *strain, b_strain);
  dolfin_set("output destination", "terminal");

  // Set up auxiliary variables for stress, strain and tangents
  double equivalent_plastic_strain(0.0);
  uBLASVector trial_stress(6), plastic_strain(6), elastic_strain(6), total_strain(6);
  uBLASDenseMatrix consistent_tangent(6,6);
  consistent_tangent.mat().clear();
  plastic_strain.vec().clear();
  elastic_strain.vec().clear();
  total_strain.vec().clear();

  // Set up arrays to get history values efficiently from function vectors
  // FIXME: I should use info from dofmaps instead, as it would be more 'safe'
  //        Could use the UFC data structure like Assembler
  double* valuesT = new double[36];
  unsigned int* rowsT = new unsigned int[36];
  double* valuesS = new double[6];
  unsigned int* rowsS = new unsigned int[6];
  double* valuesEP = new double[1];
  unsigned int* rowsEP = new unsigned int[1];

  // Get number of integration points and start loop
  // FIXME: This should doublely be a loop over elements and THEN loop all
  //        integration points locally with respect to elements.
  uint num_ips(strain_function.vector().size()/strain_function.dim(0));
  for (uint m = 0; m != num_ips; ++m)
  {
    // Elastic tangent is used in the first timestep
    if (*_elastic_tangent == true)
      consistent_tangent.mat().assign(_plastic_model->elastic_tangent.mat());

    // Consistent tangent from previous converged time step solution is used
    // as and initial guess.
    if (*_elastic_tangent != true)
    {
      for (int i = 0; i < 36; i++)
          rowsT[i] = m + i*num_ips;

      consistent_tangent_vec->get(valuesT, 36, rowsT);
      for (int i = 0; i < 6; i++)
        for (int j = 0; j < 6; j++)
          consistent_tangent.mat()(i,j) = valuesT[6*i + j];
    }

    // Get plastic strain from previous converged time step
    for (int i = 0; i < 6; i++)
      rowsS[i] = m + i*num_ips;

    plastic_strain_old->get(valuesS, 6, rowsS);
    for (int i = 0; i!=6; ++i)
      plastic_strain[i] = valuesS[i];

    // Get strains
    strain->get(valuesS, 6, rowsS);
    for (int i = 0; i < 6; i++)
      total_strain[i] = valuesS[i];

    // Compute elastic strains        
    elastic_strain.vec().assign(total_strain.vec() - plastic_strain.vec());

    // Get equivalent plastic strain from previous converged time step
    rowsEP[0] = m;
    equivalent_plastic_strain_old->get(valuesEP, 1, rowsEP);
    equivalent_plastic_strain = valuesEP[0];
      
    // Trial stresses
    trial_stress.vec().assign(prod(_plastic_model->elastic_tangent.mat(), elastic_strain.vec()));

    // Testing trial stresses, if yielding occurs the stresses are mapped 
    // back onto the yield surface, and the updated parameters are returned.
    return_mapping.ClosestPoint(*_plastic_model, consistent_tangent, trial_stress, plastic_strain, equivalent_plastic_strain);

    // Updating plastic strain 
    for (int i = 0; i < 6; i++)
      valuesS[i]= plastic_strain[i];
    plastic_strain_new->set(valuesS, 6, rowsS);

    valuesEP[0] = equivalent_plastic_strain;
    equivalent_plastic_strain_new->set(valuesEP, 1, rowsEP);

    // Update stresses for next Newton iteration
    // (trial stresses if elastic, otherwise current stress sig_c)
    // and coefficients for consistent tangent matrix
    for (int i = 0; i < 6; i++)
      valuesS[i] = trial_stress[i];

    stress->set(valuesS, 6, rowsS);

    for (int i = 0; i < 36; i++)
      rowsT[i] = m + i*num_ips;

    for (int i = 0; i < 6; i++)
      for (int j = 0; j < 6; j++)
        valuesT[6*i + j] = consistent_tangent(i,j);

    consistent_tangent_vec->set(valuesT, 36, rowsT);
  }

  delete [] valuesT; delete [] rowsT;
  delete [] valuesS; delete [] rowsS;
  delete [] valuesEP; delete [] rowsEP;
}
//-----------------------------------------------------------------------------    
// Update history variables
void PlasticityProblem::update_variables()
{
  plastic_strain_old_function = plastic_strain_new_function;
  equivalent_plastic_strain_old_function = equivalent_plastic_strain_new_function;
}
//-----------------------------------------------------------------------------
