// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#include <dolfin.h>
#include <dolfin/Plasticity.h>
#include "../../forms/3Dlinear_QE_forms/Output3D.h"

using namespace dolfin;

// Right-hand side (body force)
class Source : public Function, public TimeDependent
{
  public:

    Source(Mesh& mesh, const double* t) : Function(mesh), TimeDependent(t) {}
    void eval(double* values, const double* x) const
    {
      // Body force
      values[0] = 0.0;
      values[1] = 0.0;
      values[2] = - sin(time() / 2.0 * DOLFIN_PI)*40.0;
    }

    dolfin::uint rank() const
    {
      return 1;
    }

    dolfin::uint dim(dolfin::uint i) const
    {
      return 3;
    }

};

// Sub domain for Dirichlet boundary condition
class DirichletBoundaryXY : public SubDomain
{
  bool inside(const double* x, bool on_boundary) const
  {
    // Locate the line (0, 0, 0.5) - (0, 1, 0.5)
    return (x[0] < DOLFIN_EPS && std::abs(x[2] - 0.5) < DOLFIN_EPS);
  }
};

  // Sub domain for Dirichlet boundary condition
class DirichletBoundaryZ : public SubDomain
{
  bool inside(const double* x, bool on_boundary) const
  {
    // Locate the lines (0, 0, 0.5) - (0, 1, 0.5) and (5, 0, 0.5) - (5, 1, 0.5)
    return (x[0] < DOLFIN_EPS && std::abs(x[2] - 0.5) < DOLFIN_EPS) || (std::abs(x[0] - 5.0) < DOLFIN_EPS && std::abs(x[2] - 0.5) < DOLFIN_EPS);
  }
};

int main()
{
  // Load mesh
  Mesh mesh("../beam10000.xml.gz");
//  UnitCube mesh(1,1,1);

  // Young's modulus and Poisson's ratio
  double E = 20000.0;
  double nu = 0.3;

  // Final time and time step
  double T = 0.5;
  double dt = 0.1;
  double t = 0.0;

  // Source term, RHS
  Source f(mesh, &t);

  // Create boundary conditions
  Function zero(mesh, 0.0);
  SubSystem X(0);
  SubSystem Y(1);
  SubSystem Z(2);
  DirichletBoundaryXY dbX, dbY;
  DirichletBoundaryZ dbZ;
  DirichletBC bcX(zero, mesh, dbX, X, pointwise);
  DirichletBC bcY(zero, mesh, dbY, Y, pointwise);
  DirichletBC bcZ(zero, mesh, dbZ, Z, pointwise);

  Array<DirichletBC*> bcs(&bcX, &bcY, &bcZ);

  // Slope of hardening (linear) and hardening parameter
  double E_t(0.1 * E);
  double hardening_parameter = E_t/(1-E_t/E);

  // Yield stress
  double yield_stress = 200.0;

  // Object of class von Mise
  VonMises J2(E, nu, yield_stress, hardening_parameter);

  // Solve problem
  // Solution function
  Function u;

  // Create PlasticityProblem
  bool elastic_tangent = false;
  PlasticityProblem nonlinear_problem(u, f, mesh, bcs, elastic_tangent, J2);

  // Create nonlinear solver and set parameters
  NewtonSolver nonlinear_solver;
  nonlinear_solver.set("Newton convergence criterion", "incremental");
  nonlinear_solver.set("Newton maximum iterations", 50);
  nonlinear_solver.set("Newton relative tolerance", 1e-6);
  nonlinear_solver.set("Newton absolute tolerance", 1e-10);

  // Bilinear and linear forms for continuous output of equivalent plastic strain
  Output3DBilinearForm a_cont_equivalent_plastic_strain;
  Output3DLinearForm L_cont_equivalent_plastic_strain(nonlinear_problem.equivalent_plastic_strain_old_function);

  // Setup pde to project strains onto a continuous basis
  LinearPDE pde(a_cont_equivalent_plastic_strain, L_cont_equivalent_plastic_strain, mesh);
  pde.set("PDE linear solver", "iterative");

  // Function to hold continuous equivalent plastic strain
  Function cont_equivalent_plastic_strain;
  
  // File names for output
  File file1("output/disp.pvd");
  File file2("output/eq_plas_strain.pvd");

  dolfin::uint step_num = 0;
  while( t < T)
  {
    elastic_tangent = false;

    // Use elastic tangent in first time step
    if (t == 0.0)
      elastic_tangent = true;
    t += dt;

    // Solve non-linear problem
    nonlinear_solver.solve(nonlinear_problem, u.vector());

    // Update variables
    nonlinear_problem.update_variables();

    // Project equivalent strain onto continuous basis for postprocessing
    dolfin_set("output destination", "silent");
    pde.solve(cont_equivalent_plastic_strain);
    dolfin_set("output destination", "terminal");

    // Write output to files
    file1 << u;
    file2 << cont_equivalent_plastic_strain;

    cout << "Step num: " << ++step_num << endl;
    cout << "Time: t = " << t <<endl;
  }

  return 0;
}
