// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#include "DruckerPrager.h"

using namespace dolfin;

DruckerPrager::DruckerPrager(double E, double nu, double friction_angle, 
                double dilatancy_angle, double cohesion, double hardening_parameter) 
              : PlasticityModel(E, nu), _effective_stress(0.0), 
                _hardening_parameter(hardening_parameter)
{
  dg_dsigma_mod.vec().resize(6, false);

  alpha_friction  = (6.0 * sin(friction_angle)) / (3 - sin(friction_angle));
  k_friction      = (6.0 * cohesion * cos(friction_angle)) / (3 - sin(friction_angle));
  alpha_dilatancy = (6.0 * sin(dilatancy_angle)) / (3 - sin(dilatancy_angle));

  A.mat().resize(6, 6, false);
  A.mat().clear();
  A.mat()(0, 0) =  2; A.mat()(1, 1) =  2; A.mat()(2, 2) =  2;
  A.mat()(3, 3) =  6; A.mat()(4, 4) =  6; A.mat()(5, 5) =  6;
  A.mat()(0, 1) = -1; A.mat()(0, 2) = -1; A.mat()(1, 0) = -1;
  A.mat()(1, 2) = -1; A.mat()(2, 0) = -1; A.mat()(2, 1) = -1;    
}
//-----------------------------------------------------------------------------
double DruckerPrager::hardening_parameter(double const equivalent_plastic_strain) const
{
  return _hardening_parameter;
}
//-----------------------------------------------------------------------------
double DruckerPrager::f(const uBLASVector& stress, const double equivalent_plastic_strain)
{

  return effective_stress(stress) + alpha_friction / 3.0 * (stress[0] + stress[1] + stress[2])
         - k_friction - _hardening_parameter * equivalent_plastic_strain;
}
//-----------------------------------------------------------------------------
void DruckerPrager::df(uBLASVector& df_dsigma, const uBLASVector& stress)
{
  _effective_stress = effective_stress(stress);
  df_dsigma[0] = (2 * stress[0] - stress[1] - stress[2]) / (2 * _effective_stress) + alpha_friction / 3.0;
  df_dsigma[1] = (2 * stress[1] - stress[0] - stress[2]) / (2 * _effective_stress) + alpha_friction / 3.0;
  df_dsigma[2] = (2 * stress[2] - stress[0] - stress[1]) / (2 * _effective_stress) + alpha_friction / 3.0;
  df_dsigma[3] = 6 * stress[3] / (2 * _effective_stress); 
  df_dsigma[4] = 6 * stress[4] / (2 * _effective_stress); 
  df_dsigma[5] = 6 * stress[5] / (2 * _effective_stress);
}
//-----------------------------------------------------------------------------
void DruckerPrager::dg(uBLASVector& dg_dsigma, const uBLASVector& stress)
{
  _effective_stress = effective_stress(stress);
  dg_dsigma[0] = (2 * stress[0] - stress[1] - stress[2]) / (2 * _effective_stress) + alpha_friction / 3.0;
  dg_dsigma[1] = (2 * stress[1] - stress[0] - stress[2]) / (2 * _effective_stress) + alpha_friction / 3.0;
  dg_dsigma[2] = (2 * stress[2] - stress[0] - stress[1]) / (2 * _effective_stress) + alpha_friction / 3.0;
  dg_dsigma[3] = 6 * stress[3] / (2 * _effective_stress); 
  dg_dsigma[4] = 6 * stress[4] / (2 * _effective_stress); 
  dg_dsigma[5] = 6 * stress[5] / (2 * _effective_stress);
}
//-----------------------------------------------------------------------------
void DruckerPrager::ddg(uBLASDenseMatrix& ddg_ddsigma, const uBLASVector& stress)
{
  _effective_stress = effective_stress(stress);
  dg_mod(dg_dsigma_mod, stress);
  ddg_ddsigma.mat().assign(A.mat() / (2 * _effective_stress) - ublas::outer_prod(dg_dsigma_mod.vec(), dg_dsigma_mod.vec()) / _effective_stress);
}
//-----------------------------------------------------------------------------
double DruckerPrager::effective_stress(const uBLASVector& stress)
{
  return sqrt(std::pow((stress[0] + stress[1] + stress[2]), 2) - 3 * (stress[0] * stress[1] 
          + stress[0] * stress[2] + stress[1] * stress[2]  - std::pow(stress[3],2) 
          - std::pow(stress[4], 2)  - std::pow(stress[5], 2)));
}
//-----------------------------------------------------------------------------
void DruckerPrager::dg_mod(uBLASVector& dg_dsigma_mod, const uBLASVector& stress)
{
  _effective_stress = effective_stress(stress);
  dg_dsigma_mod[0] = (2 * stress[0] - stress[1] - stress[2]) / (2 * _effective_stress);
  dg_dsigma_mod[1] = (2 * stress[1] - stress[0] - stress[2]) / (2 * _effective_stress);
  dg_dsigma_mod[2] = (2 * stress[2] - stress[0] - stress[1]) / (2 * _effective_stress);
  dg_dsigma_mod[3] = 6 * stress[3] / (2 * _effective_stress);
  dg_dsigma_mod[4] = 6 * stress[4] / (2 * _effective_stress); 
  dg_dsigma_mod[5] = 6 * stress[5] / (2 * _effective_stress);
}
//-----------------------------------------------------------------------------

