// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#ifndef __PLASTICITY_PROBLEM_H
#define __PLASTICITY_PROBLEM_H

#include <ufc.h>
#include <dolfin/fem/Assembler.h>
#include <dolfin/fem/DirichletBC.h>
#include <dolfin/fem/UFC.h>
#include <dolfin/fem/DofMapSet.h>
#include <dolfin/function/Function.h>
#include <dolfin/nls/NonlinearProblem.h>
#include <dolfin/la/uBLASDenseMatrix.h>
#include <dolfin/la/uBLASVector.h>
#include <dolfin/la/Matrix.h>
#include <dolfin/la/Vector.h>
#include "PlasticityModel.h"
#include "ReturnMapping.h"

namespace dolfin
{

  class PlasticityProblem : public NonlinearProblem
  {
  public:

    /// Constructor
    PlasticityProblem(Function& u, Function& b, Mesh& mesh, 
         Array<DirichletBC*>& bcs, bool& elastic_tangent, PlasticityModel& plastic_model);

    /// Destructor 
    ~PlasticityProblem();

    /// Loop quadrature points and compute local tangents and stresses
    void form();

    /// User defined assemble of residual vector 
    void F(GenericVector& b, const GenericVector& x);

    /// User defined assemble of Jacobian matrix 
    void J(GenericMatrix& A, const GenericVector& x);

    /// Update variables (equivalent_plastic_strain, plastic_strain, consistent_tangent)
    void update_variables();

    /// Class variables
    // Arguments to the constructor
    Assembler assembler;
    Mesh* _mesh;
    Array<DirichletBC*>& _bcs;
    bool* _elastic_tangent;
    PlasticityModel* _plastic_model;
    bool reset_tensor;

    // Forms
    Form* a;
    Form* a_strain;
    Form* a_plastic_strain;
    Form* a_equivalent_plastic_strain;
    Form* L;
    Form* L_strain;

    // Functions
    Function consistent_tangent_function;
    Function stress_function;
    Function plastic_strain_old_function;
    Function plastic_strain_new_function;
    Function equivalent_plastic_strain_old_function;
    Function equivalent_plastic_strain_new_function;
    Function strain_function;

    // Vectors (associated with functions)
    GenericVector* consistent_tangent_vec;
    GenericVector* stress;
    GenericVector* plastic_strain_old;
    GenericVector* plastic_strain_new;
    GenericVector* equivalent_plastic_strain_old;
    GenericVector* equivalent_plastic_strain_new;
    GenericVector* strain;

    // Auxiliary variables to compute strains
    Matrix A_strain;
    Vector b_strain;

    // Object to handle return mapping of stresses back to the yield surface
    ReturnMapping return_mapping;
  };
}

#endif
