// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#include "ReturnMapping.h"
#include <dolfin/common/constants.h>
#include <boost/numeric/ublas/vector.hpp>

using namespace dolfin;

//-----------------------------------------------------------------------------
ReturnMapping::ReturnMapping(const uint maxit) : Parametrized(), maxit(maxit), plastic_flag(0)
{
  initialise();
}
//-----------------------------------------------------------------------------
ReturnMapping::~ReturnMapping()
{
  // Do nothing
}
//-----------------------------------------------------------------------------
void ReturnMapping::initialise()
{
  R.mat().resize(6, 6, false);
  ddg_ddsigma.mat().resize(6, 6, false);
  inverse_Q.mat().resize(6, 6, false);
  I.resize(6);

  df_dsigma.vec().resize(6, false);
  dg_dsigma.vec().resize(6, false);
  sigma_current.vec().resize(6, false);
  sigma_dot.vec().resize(6, false);
  sigma_residual.vec().resize(6, false);
  Rm.vec().resize(6, false);
  Rn.vec().resize(6, false);
  RinvQ.vec().resize(6, false);

  residual_f   = 0.0;
  delta_lambda = 0.0;
  lambda_dot   = 0.0;
  hardening_parameter = 0.0;
}
//-----------------------------------------------------------------------------
dolfin::uint ReturnMapping::ClosestPoint(PlasticityModel& plastic_model, uBLASDenseMatrix& consistent_tangent, 
                                 uBLASVector& trial_stresses, uBLASVector& plastic_strain, 
                                 double& equivalent_plastic_strain)
{
  // Initialise variables
  delta_lambda = 0.0;
  plastic_flag = 0;
  sigma_residual.vec().clear();
  sigma_current.vec().assign(trial_stresses.vec());

  // Compute hardening parameter
  hardening_parameter = plastic_model.hardening_parameter(equivalent_plastic_strain);

  // Evaluate yield function (trial stresses)
  residual_f = plastic_model.f(sigma_current, equivalent_plastic_strain);
//  std::cout << "f: " << residual_f << std::endl;
//  std::cout << "norm_2(sigma_current): " << norm_2(sigma_current) << std::endl;

  // Check for yielding
//  if (residual_f/ublas::norm_2(sigma_current.vec()) > DOLFIN_EPS)
  if (residual_f/ublas::norm_2(sigma_current.vec()) > 1e-12)
  {
//       std::cout << "\nif f: " << residual_f << std::endl;

    // Compute normal vectors to yield surface and plastic potential
    plastic_model.df(df_dsigma, sigma_current);
    plastic_model.dg(dg_dsigma, sigma_current);
    plastic_flag = 1;
    uint iterations(0);

    // Newton iterations to project stress onto yield surface
//    while ( std::abs(residual_f)/ublas::norm_2(sigma_current.vec()) > DOLFIN_EPS)
    while ( std::abs(residual_f)/ublas::norm_2(sigma_current.vec()) > 1e-12)
    {
      iterations++;
//      std::cout << "return mapping" << std::endl;
      if (iterations > maxit)
        error("Return mapping iterations > %d.", maxit);

      // Compute second derivative (with respect to stresses) of plastic potential
      plastic_model.ddg(ddg_ddsigma, sigma_current);
  
      // Compute auxiliary matrix Q 
      inverse_Q.mat().assign(I + delta_lambda*prod( (plastic_model.elastic_tangent.mat()), ddg_ddsigma.mat()));

      // Invert Q
      inverse_Q.invert();

      // Compute auxiliary matrix R
      R.mat().assign(ublas::prod(inverse_Q.mat(), (plastic_model.elastic_tangent.mat()) ));

      // lambda_dot, rate of plastic multiplier
      lambda_dot = (residual_f - ublas::inner_prod(sigma_residual.vec(), ublas::prod(inverse_Q.mat(), df_dsigma.vec()) ) ) / 
                        (ublas::inner_prod( df_dsigma.vec(), ublas::prod( R.mat(), dg_dsigma.vec()) ) + hardening_parameter);

//      if (lambda_dot < -DOLFIN_EPS)
//        error("lambda_dot is negative:  %g", lambda_dot);

      // Compute stress increment            
      sigma_dot.vec().assign(ublas::prod( (-lambda_dot*ublas::prod( (plastic_model.elastic_tangent.mat()), dg_dsigma.vec()) -sigma_residual.vec()), inverse_Q.mat() ));

      // Increment plastic multiplier
      delta_lambda += lambda_dot;

      // Update current stress state
      sigma_current += sigma_dot;

      // Update equivalent plastic strain
      equivalent_plastic_strain = plastic_model.kappa(equivalent_plastic_strain, sigma_current, lambda_dot);

      // Compute hardening parameter
      hardening_parameter = plastic_model.hardening_parameter(equivalent_plastic_strain);
//       std::cout << "hardening parameter: " << hardening_parameter << std::endl;
//       std::cout << "eq pl strain: " << equivalent_plastic_strain << std::endl;
//       std::cout << "sigma current: " << sigma_current << std::endl;

      // Evaluate yield function at new stress state
      residual_f = plastic_model.f(sigma_current, equivalent_plastic_strain);
//       std::cout << "f: " << residual_f << std::endl;

      // Normal to yield surface at new stress state
      plastic_model.df(df_dsigma, sigma_current);

      // Normal to plastic potential at new stress state
      plastic_model.dg(dg_dsigma, sigma_current);

      // Compute residual vector
      sigma_residual.vec().assign(sigma_current.vec() - (trial_stresses.vec() - delta_lambda*ublas::prod( (plastic_model.elastic_tangent.mat()), dg_dsigma.vec()) ) );
    }

    // Update matrices
    plastic_model.ddg(ddg_ddsigma, sigma_current);

    // Compute matrix Q
    inverse_Q.mat().assign(I + delta_lambda*ublas::prod( (plastic_model.elastic_tangent.mat()), ddg_ddsigma.mat()));

    // Invert Q
    inverse_Q.invert();

    // Compute matrix R and vector Rn
    R.mat().assign(ublas::prod(inverse_Q.mat(), (plastic_model.elastic_tangent.mat()) ));
    Rn.vec().assign(ublas::prod(R.mat(), df_dsigma.vec()));
      
    // Compute consistent tangent operator
    consistent_tangent.mat().assign(R.mat() - ublas::prod(R.mat(), ublas::outer_prod(dg_dsigma.vec(), Rn.vec()) ) / 
              (ublas::inner_prod(df_dsigma.vec(), ublas::prod(R.mat(), dg_dsigma.vec()) ) + hardening_parameter) );

    // Stresses for next Newton iteration, trial stresses are overwritten by current stresses
    trial_stresses.vec().assign(sigma_current.vec());

    // Update plastic strains
//    dg_dsigma *= delta_lambda;
//    plastic_strain += dg_dsigma;
    plastic_strain.vec() += delta_lambda * dg_dsigma.vec();
  }

  // If the trial stresses are admissible
//  if (residual_f/ublas::norm_2(sigma_current.vec()) < -DOLFIN_EPS)
  if (residual_f/ublas::norm_2(sigma_current.vec()) < -1e-12)
  {
//      std::cout << "elastic tangent" << std::endl;
      consistent_tangent.mat().assign(plastic_model.elastic_tangent.mat());
  }

  return plastic_flag;
}
//-----------------------------------------------------------------------------
