// Copyright (C) 2006-2008 Kristian Oelgaard and Garth N. Wells.
// Licensed under the GNU LGPL Version 3.
//
// First added:  2006-11-13
// Last changed: 2008-10-28

#include "VonMises.h"

using namespace dolfin;

VonMises::VonMises(double E, double nu, double yield_stress, double hardening_parameter) 
                 : PlasticityModel(E, nu), _yield_stress(yield_stress), 
                   _effective_stress(0.0), _hardening_parameter(hardening_parameter)
{
  dg_dsigma.vec().resize(6, false);

  A.mat().resize(6, 6, false);
  A.mat().clear();
  A.mat()(0, 0) =  2; A.mat()(1, 1) =  2; A.mat()(2, 2) =  2;
  A.mat()(3, 3) =  6; A.mat()(4, 4) =  6; A.mat()(5, 5) =  6;
  A.mat()(0, 1) = -1; A.mat()(0, 2) = -1; A.mat()(1, 0) = -1;
  A.mat()(1, 2) = -1; A.mat()(2, 0) = -1; A.mat()(2, 1) = -1;
}
//-----------------------------------------------------------------------------
double VonMises::hardening_parameter(const double equivalent_plastic_strain) const
{
  return _hardening_parameter;
}
//-----------------------------------------------------------------------------
double VonMises::f(const uBLASVector& stress, const double equivalent_plastic_strain)
{
  _effective_stress = effective_stress(stress);

  return _effective_stress - _yield_stress - _hardening_parameter*equivalent_plastic_strain;
}
//-----------------------------------------------------------------------------
void VonMises::df(uBLASVector& df_dsigma, const uBLASVector& stress)
{
  _effective_stress = effective_stress(stress);

  df_dsigma[0] = (2 * stress[0] - stress[1] - stress[2]) / (2 * _effective_stress);
  df_dsigma[1] = (2 * stress[1] - stress[0] - stress[2]) / (2 * _effective_stress);
  df_dsigma[2] = (2 * stress[2] - stress[0] - stress[1]) / (2 * _effective_stress);
  df_dsigma[3] = 6 * stress[3] / (2 * _effective_stress);
  df_dsigma[4] = 6 * stress[4] / (2 * _effective_stress);
  df_dsigma[5] = 6 * stress[5] / (2 * _effective_stress);
}
//-----------------------------------------------------------------------------
void VonMises::ddg(uBLASDenseMatrix& ddg_ddsigma, const uBLASVector& stress)
{
  _effective_stress = effective_stress(stress);
  df(dg_dsigma, stress);
  ddg_ddsigma.mat().assign(A.mat() / (2 * _effective_stress) - ublas::outer_prod(dg_dsigma.vec(), dg_dsigma.vec()) / _effective_stress);
}
//-----------------------------------------------------------------------------
double VonMises::effective_stress(const uBLASVector& stress)
{
  return sqrt(std::pow((stress[0] + stress[1] + stress[2]), 2)
              - 3 * (stress[0] * stress[1] + stress[0] * stress[2] + stress[1] * stress[2]
              - std::pow(stress[3], 2) - std::pow(stress[4], 2) - std::pow(stress[5], 2)));
}
//-----------------------------------------------------------------------------
